const express = require('express')
const path = require('path')
const fs = require('fs')
const crypto = require('crypto')
const cache = require('memory-cache')
const convert = require('xml-js')

const lib = require('./lib.js')

// Settings
const DIR = 'letters'
const app = express()
const PORT = 3000

function loadDocAsJson (filepath) {
  /*
   * Reads XML file and transforms it to JSON
   */

  let data = {}

  try {
    let xml = fs.readFileSync(filepath, 'utf8')
    data = convert.xml2js(xml, {compact: true})
  } catch (err) {
    console.log(arguments.callee) // eslint-disable-line no-caller
    console.log('>> Error while loading ' + filepath)
    return data
  }

  return data
}

function inCacheFormat (docname, docobj) {
  /*
   * Create some metadata for document cache representation
   */

  let hash = crypto.createHmac('sha256', JSON.stringify(docobj)).digest('hex')
  let doc = {}

  try {
    doc = {
      link: '/letters/' + docname.slice(0, -4),
      collection: docobj.TEI.teiHeader.fileDesc.titleStmt.collection._text,
      url: '/api/letters/' + docname.slice(0, -4),
      name: docname.replace(/_/g, ' ').slice(0, -4),
      hash: hash,
      object: docobj
    }
  } catch (err) {
    console.log(arguments.callee) // eslint-disable-line no-caller
    console.error('>> Error while loading ' + docname)
  }

  return doc
}

function loadDocumentsToCache (dir) {
  /*
   * Load all documents to cache
   */

  try {
    var documents = fs.readdirSync(dir)
  } catch (err) {
    console.log(arguments.callee) // eslint-disable-line no-caller
    console.log('> Error while loading files from ' + dir)
  }

  for (let docname of documents) {
    // We just want the XML documents
    if (!docname.endsWith('.xml')) { continue }
    let docjs = loadDocAsJson(path.join(dir, docname))
    cache.put(docname, inCacheFormat(docname, docjs))
  }

  console.log('> Loaded ' + cache.size() + ' documents to cache')
}

function loadDocuments (dir) {
  /*
    * Load documents from cache and update cache if necessary
    */

  // Update the cache
  // TODO: If I load the file each time, I gain nothing

  // try {
  //   var documents = fs.readdirSync(dir)
  //   var cachedocments = cache.keys()
  // } catch (err) {
  //   console.log('> Error while loading files from ' + dir)
  // }

  // // Hint: gitignore is in documents
  // if ((documents.length -1) != cachedocments.length) {
  //   console.log('> Refreshing documents cache')
  //   loadDocumentsToCache(dir)
  // }

  let data = []

  for (let item of cache.keys()) {
    data.push(cache.get(item))
  }

  return data
}

function loadDocument (doc) {
  /*
   * Load document from cache and update cache if necessary
   */

  // Update the cache
  // TODO: If I load the file each time, I gain nothing

  // var docjs = loadDocAsJson(path.join(DIR, doc))
  // const hash = crypto.createHmac('sha256', JSON.stringify(docjs)).digest('hex')

  // if (hash != cache.get(doc).hash) {
  //   console.log('> Refreshing cache for ' + doc)
  //   cache.put(doc, inCacheFormat(doc, docjs))
  // }

  return cache.get(doc)
}

// CORS Settings
app.use(function (req, res, next) {
  res.header('Access-Control-Allow-Origin', '*')
  res.header('Access-Control-Allow-Headers', 'Origin, X-Requested-With, Content-Type, Accept')
  next()
})

app.use('/', express.static('dist'))

app.get('/api/search', function (req, res) {
  // Search Endpoint with query parameters
  // <element attribute="value">content</element>
  console.time('> Search All')

  let query = {
    element: req.query.element,
    content: req.query.content
  }

  let data = []
  for (let item of cache.keys()) {
    let obj = loadDocument(item)
    let results = lib.search(obj, query)
    if (results !== null) {
      data.push(results)
    }
  }

  res.send(data)
  console.timeEnd('> Search All')
})

app.get('/api/features/:letter?', function (req, res) {
  // Feature Search Endpoint with query parameters
  // <feature category="diakritika" type="akut" subtype="akutstattgravis" ref="26">

  let query = {
    category: req.query.category,
    type: req.query.type,
    subtype: req.query.subtype
  }

  let data = []
  const DOC = req.params.letter + '.xml'

  // TODO: Refactor this shice
  if (req.params.letter) {
    console.time('> Search Features - Letter')
    let obj = loadDocument(DOC)
    let results = lib.features(obj, query)
    if (results !== null) {
      data.push(results)
    }
    console.timeEnd('> Search Features - Letter')
  } else {
    console.time('> Search Features')
    for (let item of cache.keys()) {
      let obj = loadDocument(item)
      let results = lib.features(obj, query)
      if (results !== null) {
        data.push(results)
      }
    }
    console.timeEnd('> Search Features')
  }
  res.send(data)
})

app.get('/api/letters/:letter?', function (req, res) {
  // Document Endpoint with optional Document name
  // Lists either all or one given document

  const DOC = req.params.letter + '.xml'

  if (req.params.letter) {
    console.time('> List Letter')
    res.send(loadDocument(DOC).object)
    console.timeEnd('> List Letter')
  } else {
    console.time('> List Letters')
    res.send(loadDocuments(DIR))
    console.timeEnd('> List Letters')
  }
})

// Where the magic happens
loadDocumentsToCache(DIR)

app.listen(PORT, function () {
  console.log('> Listening at http://localhost:' + PORT)
})
